import os
import pickle
from urllib.parse import urlparse
from rs4.attrdict import CaseInsensitiveDict
from aquests.protocols.http import http_util
import sys
import json

class API:
  def __init__ (self, resource_id):
    self.resource_id = resource_id
    self.endpoints = set () 
    self.routeopt = {}
    self.parameter_requirements = {}
    self.auth_requirements = []
    self.doc = None    
    self.METHODS = {} 
    self.out = None
    
  VALID_METHODS = ('GET', 'POST', 'PATCH', 'PUT', 'DELETE')
  def add_call (self, method, url, status_code, reason, reqh, reqd, resh, resd, spec):    
    if not self.endpoints:      
      self.routeopt = spec ["routeopt"]
      self.parameter_requirements = spec ["parameter_requirements"]
      self.auth_requirements = spec ["auth_requirements"]
      self.doc = spec ["doc"]
      for each in self.routeopt ['methods']:
        if each in self.VALID_METHODS:
          self.METHODS [each] = {}
    endpoint = self.get_endpoint (spec ['current_request']['uri'], spec ['routeopt']['route'])
    self.endpoints.add (endpoint)

    endpoint = self.get_endpoint (spec ['current_request']['uri'], spec ['routeopt']['route'])
    self.endpoints.add (endpoint)
    
    if method in self.METHODS:
      reqh = CaseInsensitiveDict (reqh)
      if status_code not in self.METHODS [method]:
        self.METHODS [method][status_code] = []      
      if isinstance (reqd, str):
        if reqh.get ('content-type', '').startswith ('application/json'):
          reqd = json.loads (reqd)
        else:
          reqd = http_util.crack_query (reqd)          
      
      reqd = self.squeeze (reqd)      
      resd = self.squeeze (resd)    
      for d in self.METHODS [method][status_code]:
        if d [3] == reqd: 
          return
        if set (d [3].keys ()) == set (reqd.keys ()): 
          if d [5] == resd: 
            return
        if method in ("GET", "DELETE") and d [0] == url: 
          return
      self.METHODS [method][status_code].append ((url, reason, reqh, reqd, resh, resd, spec))

  def get_endpoint (self, endpoint, route):    
    parts = urlparse (endpoint)
    a = parts.path.split ("/")
    if not route:
      return parts.path    
    b = route.split ("/")    
    if b [-1].startswith ("<path:"):
      s = parts.path.rfind ("/".join (b [:-1]))
      return parts.path [:-s] + route
    else:
      return "/".join (a [:-(len (b) - 1)] + b [1:])

  def squeeze (self, d):
    if isinstance (d, dict):
      d_ = {}
      for k, v in list (d.items ()):
        if isinstance (v, (dict, list, str)):
          d_ [k] = self.squeeze (v) 
        else:
          d_ [k] = v  
      return d_     
    elif isinstance (d, list):
      d_ = []
      for each in d:
        d_.append (self.squeeze (each))
      return d_
    elif isinstance (d, str) and len (d) > 80:
      return d [:77] + '...'    
    return d  

  def to_dict (self, d):
    return crack_query.crack_query (d)    

  def write (self, d):
    self.out and self.out.write (d)
    sys.stdout.write (d)

  def writeln (self, line = ''):
    self.write (line + '  \n')

  def writebl (self, data, code = 'json'):
    self.write ("\n```{}\n{}\n```\n\n".format (code, data))

  def writep (self, line = ''):
    self.write (line + '\n\n')
  
  def chiplist (self, l):
    return "`{}`".format ("`, `".join (sorted (l)))

  def argslist (self, part):
    if self.parameter_requirements.get (part):
      self.writeln ("**Required {} Parameters**:".format (part))
      for arg in self.parameter_requirements [part][0]:
        self.writeln ("  - {}".format (arg))
        
  def render (self, out):    
    self.out = out
    self.writep ('# {}'.format (self.resource_id))
    if self.doc:
      self.writep (self.doc)
    
    self.writeln ('**Endpoint(s)**: {}'.format (self.chiplist (list (self.endpoints))))
    self.writep ("**Methods Allowed**: {}".format (self.chiplist (self.routeopt ['methods'])))

    if self.routeopt.get ('urlargs'):
      self.writeln ("**URL Parameters**: {}".format (self.chiplist (self.routeopt ['args'][:self.routeopt ['urlargs']])))

    qss = self.parameter_requirements.get ("ARGS", [[]]) [0]
    if not qss:
      qss = self.routeopt ['args'][self.routeopt.get ('urlargs', 0):]
    if qss:
      self.writeln ("**Method-Wide Common Parameters**:")
      defaults = self.routeopt.get ('defaults', {})
      for arg in qss:
        if arg in defaults:
          if defaults [arg]:
            need = "optional, default: {}".format (defaults [arg])
          else:
            need = "optional".format ()  
        else:
          need = "required"  
        self.writeln ("  - {} (*{}*)".format (arg, need))
    self.writeln ()

    auth = 'NO'
    perms = []
    testpass = None
    login = False
    for t, d in self.auth_requirements:
      if t == "permission":
        if auth == 'NO':
          auth = "YES"
        perms = d or []
      elif t == "login":
        login = True
      elif t == "testpass":
        testpass = d
      elif t == "auth":  
        auth = d

    auth and self.writeln ("**Authorization Required**: {}".format (auth))
    perms and self.writeln ("**Permission Required**: {}".format (self.chiplist (perms)))
    testpass and self.writeln ("**Test Pass Required**: `{}`".format (testpass))
    login and self.writeln ("**Login Required**: YES")
    self.writeln ()    
    #{'methods': ['OPTIONS', 'GET', 'DELETE'], 'route': '/<name>', 'args': ['name', 'where'], 'defaults': {'where': 'latest'}, 'urlargs': 1}    
    for method in self.VALID_METHODS:
      if method not in self.METHODS:
        continue
      if not self.METHODS [method]:
        continue

      self.writep ("### {}".format (method))
      if method in ("POST", "PUT", "PATCH"):
        self.argslist ('JSON')
        self.argslist ('FORM')      
      else:
        self.argslist ('URL')
      self.writeln ()

      idx = [0, 0]
      for status_code in sorted (self.METHODS [method].keys ()):
        for url, reason, reqh, reqd, resh, resd, spec in self.METHODS [method][status_code]:          
          if 200 <= status_code < 300:
            idx [0] += 1
            if idx [0] >= 3:
              continue
            self.writep ("**Sucess Response Example**")
            self.writeln ("> **_URL_**: `{}`".format (spec ["current_request"]['uri']))
            if reqd:              
              self.writeln ("> **_Request Data_** ({})".format (reqh.get ('content-type')))
              self.writebl (json.dumps (reqd, ensure_ascii = False, indent = 2))
            self.writeln ()
            self.writeln ("> **_Response_**: `{} {}`".format (status_code, reason))
            if resd:
              self.writeln ("> **_Response Data_** ({})".format (resh.get ('content-type')))
              self.writebl (json.dumps (resd, ensure_ascii = False, indent = 2))          

          else:
            idx [1] += 1
            self.writep ("**Error Response Example**")
            self.writeln ("> **_Response_**: `{} {}`".format (status_code, reason))
            if resh.get ('content-type'):
              self.writeln ("> **_Response Data_** ({})".format (resh.get ('content-type')))    
            if resd:
              self.writebl (json.dumps (resd, ensure_ascii = False, indent = 2))              
          
          self.writeln ()

apis = {}
def build_doc ():
  for fn in os.listdir (".webtest_log/v"):
    with open (os.path.join (".webtest_log/v", fn), 'rb') as f:
      method, url, status_code, reason, reqh, reqd, resh, resd, spec = pickle.load (f)

    if not spec:
      continue
    resource_id = spec ['id']  
    if resource_id not in apis:
      apis [resource_id] = API (resource_id)
    apis [resource_id].add_call (method.upper (), url, status_code, reason, reqh, reqd, resh, resd, spec)
  
  methods_count = {}
  with open ("../README.md", 'w') as out:
  #with sys.stdout as out:
    out.write ("## Table of Content\n\n")
    sorted_apis = sorted (apis.items (), key = lambda x: x [0])
    for idx, (resource_id, api) in enumerate (sorted_apis):     
      end_points = " | ".join (list (api.endpoints)).replace ("<", "&lt;").replace (">", "&gt;")
      out.write ("1. [{}](#{})".format (end_points, resource_id.replace (".", "").lower ()))
      for method in API.VALID_METHODS:
        if not api.METHODS.get (method):
          continue        
        try: 
          methods_count [method] += 1
        except KeyError: 
          methods_count [method] = 0
          out.write (" [**`{}`**](#{})".format (method, method.lower ()))  
        else:  
          out.write (" [**`{}`**](#{}-{}) ".format (method, method.lower (), methods_count [method]))  
        
      out.write ("\n")

    for resource_id, api in sorted_apis:    
      api.render (out)
      out.write ("\n\n")

# logging spec -----------------------------------------------

def truncate_log_dir (remove_only = False):
  from . import pathtool
  import shutil

  if os.path.isdir (".webtest_log/v"):
      shutil.rmtree (".webtest_log/v")
  if remove_only:
      return     
  pathtool.mkdir (".webtest_log/v")

SPECS = 0
if os.path.isdir (".webtest_log/v"):
  for fn in os.listdir (".webtest_log/v"):
    SPECS = max (SPECS, int (fn [:-5]))

def log_spec (method, url, status_code, reason, reqh, reqd, resh, resd):
    global SPECS

    try:
        spec = resd.pop ("__spec__")
    except (AttributeError, KeyError):
        return
    if not os.path.isdir (".webtest_log/v"):
      return
    if SPECS > 1000:
      truncate_log_dir ()
      SPECS = 0    
    SPECS += 1
    with open (os.path.join (".webtest_log/v", '{:04d}.spec'.format (SPECS)), 'wb') as f:
        pickle.dump ((method, url, status_code, reason, reqh, reqd, resh, resd, spec), f)
       





import os

# POSIX PROCESS LOCK

if os.name == "posix":
	import fcntl
	
	class Lock:	
		def __init__(self, filename):
			self.filename = filename
			# This will create it if it does not exist already
			self.handle = open (filename, 'w')
		
		# Bitwise OR fcntl.LOCK_NB if you need a non-blocking lock 
		def acquire(self):
			fcntl.flock(self.handle, fcntl.LOCK_EX)
			
		def release(self):
			fcntl.flock(self.handle, fcntl.LOCK_UN)
		
		def close (self):
			self.handle.close()			
				
		def __del__(self):
			self.close ()
			
		def __enter__ (self):
			self.acquire ()
			return self
	
		def __exit__ (self, type, value, tb):
			self.release ()

else:
	import multiprocessing
	
	Lock = multiprocessing.RLock ()
	



import os, sys
try:
    from setproctitle import setproctitle
except ImportError:
    setproctitle = None

from . import daemon
import time
from .. import pathtool
from ..termcolor import tc
if os.name == "nt":
    import win32serviceutil
    
class Service:
    def __init__ (self, name, working_dir, lockpath = None, win32service = None):
        self.name = name
        self.working_dir = working_dir
        self.lockpath = lockpath or working_dir
        self.win32service = win32service        
        setproctitle and setproctitle (name)
        pathtool.mkdir (working_dir)
        if lockpath:
            pathtool.mkdir (lockpath)
        os.chdir (working_dir)
        
    def start (self):
        if os.name == "nt":
            set_service_config (['start'])
        else:    
            from . import Daemonizer            
            if not Daemonizer (self.working_dir, self.name, lockpath = self.lockpath).runAsDaemon ():
                pid = daemon.status (self.lockpath, self.name)                
                print ("{} {}".format (tc.debug (self.name), tc.error ("[already running:{}]".format (pid))))
                sys.exit ()
            
    def stop (self):
        if os.name == "nt":            
            set_service_config (['stop'])
        else:    
            daemon.kill (self.lockpath, self.name, True)
    
    def status (self, verbose = True):
        pid = daemon.status (self.lockpath, self.name)
        if verbose:
            if pid:
                print ("{} {}".format (tc.debug (self.name), tc.warn ("[running:{}]".format (pid))))
            else:
                print ("{} {}".format (tc.debug (self.name), tc.secondary ("[stopped]")))
        return pid
    
    def execute (self, cmd):
        if cmd == "stop":
            self.stop ()
            return False
        elif cmd == "status":
            self.status ()
            return False
        elif cmd == "start":
            self.start ()
        elif cmd == "restart":
            self.stop ()        
            time.sleep (2)
            self.start ()
        elif cmd == "install":    
            self.install ()
        elif cmd == "update":    
            self.update ()
        elif cmd == "remove":    
            self.remove ()
        else:
            raise AssertionError ('unknown command: %s' % cmd)
                
        if os.name == "nt":
            return False
        return True
    
    if os.name == "nt":
        def set_service_config (self, argv = []):
            argv.insert (0, "")            
            script = os.path.join (os.getcwd (), sys.argv [0])
            win32serviceutil.HandleCommandLine(self.win32service, "%s.%s" % (script [:-3], self.win32service.__name__), argv)
                
        def install (self):
            self.set_service_config (['--startup', 'auto', 'install'])
    
        def remove (self):
            self.set_service_config (['remove'])
        
        def update (self):
            self.set_service_config (['update'])


            


# pytest framework ---------------------------------------------
import requests
from rs4 import siesta
import time
import sys
import os
import xmlrpc.client 
from io import IOBase
import json
from . import apidoc

class Target:
    # f = Target ('http://localhost:')
    # f.get ("/v1/accounts/me")
    # == f.http.get ("/v1/accounts/me")
    # f.axios.get ("/v1/accounts/me")    
    # f.siesta.v1.accounts ("me").get ()

    def __init__ (self, endpoint, api_call = False):
        self.endpoint = endpoint        
        self.s = requests.Session ()
        self._api_call = api_call
        self.siesta = siesta.API (endpoint, reraise_http_error = False)
        if not self._api_call:
            self.axios = Target (endpoint, True)
        else:
            self.set_default_header ('Accept', "application/json")        
            self.set_default_header ('Content-Type', "application/json")
        
    @property
    def http (self):
        return self

    def set_default_header (self, k, v):
        self.s.headers.update ({k: v})

    def api (self, point = None):
        if point:
            return siesta.API (point, reraise_http_error = False)
        return self.siesta

    def __enter__ (self):
        return self
        
    def __exit__ (self, type, value, tb):
        self._close ()
        
    def __del__ (self):
        self._close ()
            
    def _close (self):
        pass

    def resolve (self, url):
        if url.startswith ("http://") or url.startswith ("https://"):
            return url
        else:
            return self.endpoint + url 

    def _request (self, method, url, *args, **kargs):
        url = self.resolve (url)
        rf = getattr (self.s, method)
        if args:
            args = list (args)
            request_data = args.pop (0)
            args = tuple (args)
        else:
            try:
                request_data = kargs.pop ('data')
            except KeyError:
                request_data = None    

        if isinstance (request_data, dict):
            request_data = json.dumps (request_data)
        
        if request_data:   
            resp = rf (url, request_data, *args, **kargs)
        else:
            resp = rf (url, *args, **kargs)

        if resp.headers.get ('content-type', '').startswith ('application/json'):            
            resp.data = resp.json ()
            if "__spec__" in resp.data:
                reqh = kargs.get ('headers', {})
                reqh.update (self.s.headers)                
                apidoc.log_spec (method.upper (), url, resp.status_code, resp.reason, reqh, request_data, resp.headers, resp.data)            
        else:
            resp.data = resp.content    
        return resp

    def get (self, url, *args, **karg):
        return self._request ('get', url, *args, **karg)
        
    def post (self, url, *args, **karg):
        return self._request ('post', url, *args, **karg)
    
    def upload (self, url, data, **karg):
        files = {}
        for k in list (data.keys ()):
            if isinstance (data [k], IOBase):
                files [k] = data.pop (k)        
        return self._request ('post', url, files = files, data = data, **karg)

    def put (self, url, *args, **karg):
        return self._request ('put', url, *args, **karg)
    
    def patch (self, url, *args, **karg):
        return self._request ('patch', url, *args, **karg)
    
    def delete (self, url, *args, **karg):
        return self._request ('delete', url, *args, **karg)
    
    def head (self, url, *args, **karg):
        return self._request ('head', url, *args, **karg)
                
    def options (self, url, *args, **karg):
        return self._request ('options', url, *args, **karg)
    
    def rpc (self, url, proxy_class = None):
        return (proxy_class or xmlrpc.client.ServerProxy) (self.resolve (url))
    xmlrpc = rpc
    
    def jsonrpc (self, url, proxy_class = None):
        import jsonrpclib
        return (proxy_class or jsonrpclib.ServerProxy) (self.resolve (url))
    
    def grpc (self, url, proxy_class = None):
        raise NotImplementedError

if __name__ == "__main__":
    if "--gendoc" in sys.argv:
        apidoc.build_doc ()
    elif "--logon" in sys.argv:
        apidoc.truncate_log_dir ()
    elif "--logoff" in sys.argv:
        apidoc.truncate_log_dir (remove_only = True)    
    




# Copyright (C) 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for the Google Assistant API."""

import logging

from google.assistant.embedded.v1alpha2 import embedded_assistant_pb2


def log_assist_request_without_audio(assist_request):
    """Log AssistRequest fields without audio data."""
    if logging.getLogger().isEnabledFor(logging.DEBUG):
        resp_copy = embedded_assistant_pb2.AssistRequest()
        resp_copy.CopyFrom(assist_request)
        if len(resp_copy.audio_in) > 0:
            size = len(resp_copy.audio_in)
            resp_copy.ClearField('audio_in')
            logging.debug('AssistRequest: audio_in (%d bytes)',
                          size)
            return
        logging.debug('AssistRequest: %s', resp_copy)


def log_assist_response_without_audio(assist_response):
    """Log AssistResponse fields without audio data."""
    if logging.getLogger().isEnabledFor(logging.DEBUG):
        resp_copy = embedded_assistant_pb2.AssistResponse()
        resp_copy.CopyFrom(assist_response)
        has_audio_data = (resp_copy.HasField('audio_out') and
                          len(resp_copy.audio_out.audio_data) > 0)
        if has_audio_data:
            size = len(resp_copy.audio_out.audio_data)
            resp_copy.audio_out.ClearField('audio_data')
            if resp_copy.audio_out.ListFields():
                logging.debug('AssistResponse: %s audio_data (%d bytes)',
                              resp_copy,
                              size)
            else:
                logging.debug('AssistResponse: audio_data (%d bytes)',
                              size)
            return
        logging.debug('AssistResponse: %s', resp_copy)



import fcntl, socket, struct
import subprocess
import os

def get_mac (iname):
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    info = fcntl.ioctl(s.fileno(), 0x8927,  struct.pack('256s', iname [:15].encode ("utf8")))
    return ':'.join(['%02x' % hex for hex in info[18:24]])

def get_ip ():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    try:
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP

def get_hwid():
    if 'nt' in os.name:
        return subprocess.Popen('dmidecode.exe -s system-uuid'.split())
    else:
        return subprocess.Popen('hal-get-property --udi /org/freedesktop/Hal/devices/computer --key system.hardware.uuid'.split())
        
        


import os

if os.name == "nt":
    from colorama import init
    init()

class tc:
    WHITE = '\033[97m'
    CYAN = '\033[96m'
    MAGENTA = '\033[95m'
    OKBLUE = '\033[94m'
    WARNING = '\033[93m'
    OKGREEN = '\033[92m'
    FAIL = '\033[91m'
    GREY = '\033[90m'
    
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'
    ENDC = '\033[0m'
    
    @classmethod
    def _wrap (cls, s, c):
        return "{}{}{}".format (c, s, cls.ENDC)
    
    @classmethod
    def default (cls, s):
        return s
    
    @classmethod
    def critical (cls, s):
        return cls._wrap (s, cls.GREY)

    @classmethod
    def fatal (cls, s):
        return cls._wrap (s, cls.GREY)

    @classmethod
    def info (cls, s):
        return cls._wrap (s, cls.OKGREEN)
    
    @classmethod
    def warn (cls, s):
        return cls._wrap (s, cls.WARNING)
    
    @classmethod
    def expt (cls, s):
        return cls._wrap (s, cls.MAGENTA)
    
    @classmethod
    def fail (cls, s):
        return cls._wrap (s, cls.CYAN)
    
    @classmethod
    def error (cls, s):
        return cls._wrap (s, cls.FAIL)
    
    @classmethod
    def debug (cls, s):
        return cls._wrap (s, cls.WHITE)
            
    @classmethod
    def echo (cls, s):
        return cls._wrap (s, cls.WHITE)
    
    @classmethod
    def primary (cls, s):
        return cls._wrap (s, cls.OKBLUE)
    
    @classmethod
    def secondary (cls, s): 
        return cls._wrap (s, cls.CYAN)
    


import os
import re
import glob
import types
import sys
from setuptools.command.easy_install import easy_install

try:
	import urllib.parse
	unquote = urllib.parse.unquote
except ImportError:
	import urlparse
	unquote = urlparse.unquote

def remove (d):
	targets = glob.glob (d)
	if not targets: return
	
	for path in targets:
		if os.path.isdir (path):
			remove_silent (os.path.join (path, "*"))
			try: os.rmdir (path)
			except: pass
			continue			
		os.remove (path)
			
def mkdir (tdir, mod = 0o755):
	tdir = os.path.abspath (tdir)
	while tdir:
		if tdir [-1] in ("\\/"):
			tdir = tdir [:-1]
		else:
			break	

	if os.path.isdir (tdir): return	
	chain = [tdir]	
	while 1:
		tdir, last = os.path.split (tdir)		
		if not last: 
			break
		if tdir:
			chain.insert (0, tdir)
	
	for dir in chain [1:]:
		if os.path.isdir (dir):
			continue
		os.mkdir (dir)
		if os.name == "posix":
			os.chmod (dir, mod)
			
def modpath (mod_name):	
	if type (mod_name) in (str, bytes):		
		try:
				mod = sys.modules [mod_name]
		except KeyError: 
			return "", ""		
		return mod.__name__, mod.__file__

class easy_install_default(easy_install):
	def __init__(self):
		from distutils.dist import Distribution		
		self.distribution = Distribution()
		self.initialize_options()
		
def get_package_dir ():
	e = easy_install_default()
	try: e.finalize_options()
	except: pass
	return e.install_dir
	
				
NAFN = re.compile (r"[\\/:*?\"<>|]+")
def mkfn (text):
	text = unquote (text)
	return NAFN.sub ("_", text)




# backward compatible
from .apis.google import *

import warnings

warnings.simplefilter('default')
warnings.warn (
   "rs.googleapi will be deprecated, use rs3.apis.google",
    DeprecationWarning
)



import getopt as optparse
import sys

long_opts = []
short_opts = []

def add_option (lname = None, sname = None):
    global long_opts, short_opts    
    if lname:
        if lname.startswith ("--"):
            lname = lname [2:]
        long_opts.append (lname)

    if sname:
        if sname.startswith ("-"):
            sname = sname [1:]        
        short_opts.append (sname)

def add_options (*names):
    for name in names:
        assert name and name [0] == "-"
        if name.startswith ("--"):
            add_option (name [2:])
        else:
            add_option (None, name [1:])

class ArgumentOptions:
    def __init__ (self, kopt = {}, argv = []):
        self.__kopt = kopt
        self.argv = argv

    def items (self):
        return self.__kopt.items ()

    def __contains__ (self, k):
        return k in self.__kopt

    def get (self, k, v = None):
        return self.__kopt.get (k, v)

    def set (self, k, v = None):
        self.__kopt [k] = v

    def remove (self, k):
        try:
            del self.__kopt [k]
        except KeyError:
            pass  

def options ():
    global long_opts, short_opts
    
    argopt = optparse.getopt (sys.argv [1:], "".join (short_opts).replace ("=", ":"), long_opts)
    kopts_ = {}
    for k, v in argopt [0]:
        if k in kopts_:
            if not isinstance (kopts_ [k], list):
                kopts [k] = {kopts [k]}
            kopts_ [k].add (v)
        else:
            kopts_ [k] = v
    return ArgumentOptions (kopts_, argopt [1])




# Copyright (C) 2017 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Helper functions for the Device Actions."""

import concurrent.futures
import logging
import sys
import json
import google.auth.transport.requests
import click
import os

def register (api_endpoint, project_id, device_model_id, device_config = None):
    device_config = device_config or os.path.join (click.get_app_dir('googlesamples-assistant'), 'device_config.json')
    try:
        with open(device_config) as f:
            device = json.load(f)
            device_id = device['id']
            device_model_id = device['model_id']
            logging.info("Using device model %s and device id %s",
                            device_model_id,
                            device_id)
        
    except Exception as e:
        logging.warning('Device config not found: %s' % e)
        logging.info('Registering device')
        if not device_model_id:
            logging.error('Option --device-model-id required '
                            'when registering a device instance.')
            sys.exit(-1)
        if not project_id:
            logging.error('Option --project-id required '
                            'when registering a device instance.')
            sys.exit(-1)
        device_base_url = (
            'https://%s/v1alpha2/projects/%s/devices' % (api_endpoint,
                                                            project_id)
        )
        device_id = str(uuid.uuid1())
        payload = {
            'id': device_id,
            'model_id': device_model_id,
            'client_type': 'SDK_SERVICE'
        }
        session = google.auth.transport.requests.AuthorizedSession(
            credentials
        )
        r = session.post(device_base_url, data=json.dumps(payload))
        if r.status_code != 200:
            logging.error('Failed to register device: %s', r.text)
            sys.exit(-1)
        logging.info('Device registered: %s', device_id)
        pathlib.Path(os.path.dirname(device_config)).mkdir(exist_ok=True)
        with open(device_config, 'w') as f:
            json.dump(payload, f)
    
    return device_id, device_model_id
    


key_inputs_ = 'inputs'
key_intent_ = 'intent'
key_payload_ = 'payload'
key_commands_ = 'commands'
key_id_ = 'id'


class DeviceRequestHandler(object):
    """Asynchronous dispatcher for Device actions commands.

    Dispatch commands to the given device handlers.

    Args:
      device_id: device id to match command against

    Example:
      # Use as as decorator to register handler.
      device_handler = DeviceRequestHandler('my-device')
      @device_handler.command('INTENT_NAME')
      def handler(param):
          pass
    """

    def __init__(self, device_id):
        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        self.device_id = device_id
        self.handlers = {}

    def __call__(self, device_request):
        """Handle incoming device request.

        Returns: List of concurrent.futures for each command execution.
        """
        fs = []
        if key_inputs_ in device_request:
            for input in device_request[key_inputs_]:
                if input[key_intent_] == 'action.devices.EXECUTE':
                    for command in input[key_payload_][key_commands_]:
                        fs.extend(self.submit_commands(**command))
        return fs

    def command(self, intent):
        """Register a device action handlers."""
        def decorator(fn):
            self.handlers[intent] = fn
        return decorator

    def submit_commands(self, devices, execution):
        """Submit device command executions.

        Returns: a list of concurrent.futures for scheduled executions.
        """
        fs = []
        for device in devices:
            if device[key_id_] != self.device_id:
                logging.warning('Ignoring command for unknown device: %s'
                                % device[key_id_])
                continue
            if not execution:
                logging.warning('Ignoring noop execution')
                continue
            for command in execution:
                f = self.executor.submit(
                    self.dispatch_command, **command
                )
                fs.append(f)
        return fs

    def dispatch_command(self, command, params=None):
        """Dispatch device commands to the appropriate handler."""
        try:
            if command in self.handlers:
                self.handlers[command](**params)
            else:
                logging.warning('Unsupported command: %s: %s',
                                command, params)
        except Exception as e:
            logging.warning('Error during command execution',
                            exc_info=sys.exc_info())
            raise e




import warnings
from functools import wraps

def deprecated (msg = ""):
    def decorator (f):
        @wraps(f)
        def wrapper (was, *args, **kwargs):
            warnings.simplefilter ('default')
            warnings.warn (
               "{} will be deprecated{}".format (f.__name__, msg and (", " + msg) or ""),
                DeprecationWarning
            )
            return f (was, *args, **kwargs)
        return wrapper
    return decorator



"""
Hans Roh 2015 -- http://osp.skitai.com
License: BSD
"""
import re
import sys
import os
import shutil, glob
from warnings import warn
try:
	from setuptools import setup
except ImportError:
	from distutils.core import setup

with open('rs4/__init__.py', 'r') as fd:
	version = re.search(r'^__version__\s*=\s*"(.*?)"',fd.read(), re.M).group(1)
	
if sys.argv[-1] == 'publish':
	buildopt = ['sdist', 'upload']
	if os.name == "nt":
		buildopt.insert (0, 'bdist_wheel')		
	os.system('python setup.py %s' % " ".join (buildopt))
	#os.system('twine upload dist/rs-%s*' % version)
	for each in os.listdir ("dist"):
		os.remove (os.path.join ('dist', each))
	sys.exit()

elif sys.argv[-1] == 'link':
	import site
	if os.name == "nt":
		linkdir = [each for each in site.getsitepackages() if each.endswith ("-packages")][0]		
	else:
		linkdir = [each for each in site.getsitepackages() if each.find ("/local/") !=- 1 and each.endswith ("-packages")][0]		
	target = os.path.join (os.path.join (os.getcwd (), os.path.dirname (__file__)), "rs4")
	link = os.path.join (linkdir, "rs4")
	if os.name == "nt":
		os.system ("mklink /d {} {}".format (link, target))
	else:		
		os.system ("ln -s {} {}".format (target, link))	
	sys.exit ()
	
classifiers = [
  'License :: OSI Approved :: MIT License',
  'Development Status :: 4 - Beta',  
  'Topic :: Utilities',
	'Environment :: Console',	
	'Topic :: Software Development :: Libraries :: Python Modules',
	'Intended Audience :: Developers',	
	'Programming Language :: Python',	
	'Programming Language :: Python :: 3'
]

PY_MAJOR_VERSION = sys.version_info [0]
if PY_MAJOR_VERSION == 3:
	if os.path.isfile ("rs4/lib/py2utils.py"):
		os.remove ("rs4/lib/py2utils.py")
		try: os.remove ("rs4/lib/py2utils.pyc")
		except OSError: pass		
else:
	if not os.path.isfile ("rs4/lib/py2utils.py"):
		with open ("rs4/lib/py2utils.py", "w") as f:
			f.write ("def reraise(type, value, tb):\n\traise type, value, tb\n")			

packages = [
	'rs4',
  	'rs4.psutil',
	'rs4.webkit',
	'rs4.nets',
	'rs4.nets.ipsec',
	'rs4.nets.proxy',
	'rs4.apis',
	'rs4.apis.aws',
	'rs4.apis.aws.ec2',
	'rs4.apis.google',
	'rs4.apis.google.assistant',
	'rs4.apis.etri',
	'rs4.apis.houndify'
]

package_dir = {'rs4': 'rs4'}
package_data = {}

install_requires = [	
	"psutil",	
	"html2text", # need from siesta
	"requests",
	"colorama",
	"tqdm",
	"event_bus==1.0.2"
]

if os.name == "posix":
	install_requires.append ("setproctitle")
	install_requires.append ("ujson")

with open ('README.rst', encoding='utf-8') as f:
	long_description = f.read()
	
setup(
	name='rs4',
	version=version,
	description='Utility Pack',
	long_description = long_description,
	url = 'https://gitlab.com/hansroh/rs4',
	author='Hans Roh',	
	author_email='hansroh@gmail.com',	
	packages=packages,
	package_dir=package_dir,
	package_data = package_data,
	license='MIT',
	platforms = ["posix", "nt"],
	download_url = "https://pypi.python.org/pypi/rs4",
	install_requires = install_requires,
	classifiers=classifiers
)
